##################################################
# SPYCE - Python-based HTML Scripting
# Copyright (c) 2002 Rimon Barr.
#
# Refer to spyce.py
# CVS: $Id: form.py,v 1.1 2003/11/26 22:52:59 batripler Exp $
##################################################

__doc__ = '''Spyce tags for create data-populated forms.'''

from spyceTag import spyceTagLibrary, spyceTagPlus, spyceTagSyntaxException
import string

def formatArgs(kwargs):
  s = []
  for key in kwargs.keys():
    if kwargs[key]==None:
      s.append( key )
    else:
      s.append( '%s="%s"' % (key, kwargs[key]) )
  s = string.join(s, ' ')
  if s: s = ' '+s
  return s

def _getFormElement(tag, name, attr):
  tag = tag.getParent('form')
  if tag==None: return
  if not hasattr(tag, attr): return
  attr = getattr(tag, attr)
  if attr==None: return
  if not attr.has_key(name): return
  return attr[name]

def getFormValue(tag, name):
  return _getFormElement(tag, name, 'value')

def getFormDefault(tag, name):
  return _getFormElement(tag, name, 'default')

class form_form(spyceTagPlus):
  name = 'form'
  mustend = 1
  def syntax(self):
    self.syntaxPairOnly()
  def begin(self, method=None, action=None, value=None, default=None, **kwargs):
    if method==None:
      method = 'GET'
    else:
      method = self.contextEval(method)
    method = string.upper(method)
    if method not in ['GET', 'POST']:
      raise 'invalid method attribute value: '+method
    if action==None:
      action = self.contextGetModule('request').uri_path()
    else:
      action = self.contextEval(action)
    self.value = self.contextEval(value)
    self.default = self.contextEval(default)
    self.getOut().write('<form method="%s" action="%s"%s>' % (
      method, action, formatArgs(kwargs)) )
  def end(self):
    self.getOut().write('</form>')

class form_submit(spyceTagPlus):
  name = 'submit'
  def syntax(self):
    self.syntaxSingleOnly()
  def begin(self, name=None, value=None, **kwargs):
    if name!=None:
      kwargs['name'] = self.contextEval(name)
    if value!=None:
      kwargs['value'] = self.contextEval(value)
    self.getOut().write('<input type="SUBMIT"%s>' % (
      formatArgs(kwargs),) )

class form_hidden(spyceTagPlus):
  name = 'hidden'
  def syntax(self):
    self.syntaxSingleOnly()
    self.syntaxExist('name')
  def begin(self, name, value=None, default=None, **kwargs):
    name = self.contextEval(name)
    value = self.contextEval(value)
    if value==None:
      value = getFormValue(self, name)
      if value==None:
        default = self.contextEval(default)
        if default==None:
          default = getFormDefault(self, name)
        value = self.contextGetModule('request').getpost1(name, default)
        if value==None:
          value = ''
    self.getOut().write('<input type="%s" name="%s" value="%s"%s>' % (
      self.name, name, value, formatArgs(kwargs)) )

class form_text(form_hidden):
  name = 'text'
  def begin(self, name, value=None, default=None, size=None, maxlength=None, **kwargs):
    if size!=None:
      kwargs['size'] = self.contextEval(size)
    if maxlength!=None:
      kwargs['maxlength'] = self.contextEval(maxlength)
    apply(form_hidden.begin, (self, name, value, default), kwargs)

class form_password(form_text):
  name = 'password'

class form_textarea(spyceTagPlus):
  name = 'textarea'
  buffer = 1
  def syntax(self):
    self.syntaxPairOnly()
    self.syntaxExist('name')
  def begin(self, name, value=None, rows=None, cols=None, **kwargs):
    self._name = self.contextEval(name)
    self._value = self.contextEval(value)
    if self._value==None:
      self._value = getFormValue(self, self._name)
    if rows!=None:
      kwargs['rows'] = self.contextEval(rows)
    if cols!=None:
      kwargs['cols'] = self.contextEval(cols)
    self._args = kwargs
  def body(self, content):
    if self._value==None:
      if content==None or not string.strip(content):
        content = getFormDefault(self, self._name)
      self._value = self.contextGetModule('request').getpost1(self._name, content)
      if self._value==None:
        self._value = ''
    self.getOut().write('<textarea name="%s" %s>%s</textarea>' % (
      self._name, formatArgs(self._args), self._value) )

class form_radio(spyceTagPlus):
  name = 'radio'
  def syntax(self):
    self.syntaxSingleOnly()
    self.syntaxExist('name', 'value')
  def begin(self, name, value, checked=0, default=0, **kwargs):
    name = self.contextEval(name)
    value = self.contextEval(value)
    if checked==None: checked=1
    if default==None: default=1
    if not checked:
      checkedValues = getFormValue(self, name)
      if checkedValues==None:
        checkedValues = self.contextGetModule('request').getpost(name)
      if checkedValues!=None and type(checkedValues) not in [type([]), type(())]:
        checkedValues = [checkedValues,]
      if checkedValues!=None:
        checked = value in checkedValues
      else:
        if default:
          checked = 1
        else:
          defaultValues = getFormDefault(self, name)
          if defaultValues!=None and type(defaultValues) not in [type([]), type(())]:
            defaultValues = [defaultValues,]
          if defaultValues!=None:
            checked = value in defaultValues
    if checked:
      checkedstr = ' CHECKED'
    else:
      checkedstr = ''
    self.getOut().write('<input type="%s" name="%s" value="%s"%s%s>' % (
      self.name, name, value, checkedstr, formatArgs(kwargs)) )

class form_checkbox(form_radio):
  name = 'checkbox'

class form_select(spyceTagPlus):
  name = 'select'
  def syntax(self):
    self.syntaxPairOnly()
    self.syntaxExist('name')
  def begin(self, name, value=None, default=None, multiple=0, size=None, **kwargs):
    self.varname = self.contextEval(name)
    self.values = self.contextEval(value)
    self.default = self.contextEval(default)
    multiple = self.contextEval(multiple)
    if multiple==None: multiple = 1
    if multiple:
      multiplestr = ' MULTIPLE'
    else:
      multiplestr = ''
    size = self.contextEval(size)
    if size==None:
      sizestr = ''
    else:
      sizestr = ' size="%s"' % size
    self.getOut().write('<select name="%s"%s%s%s>' % (
      self.varname, sizestr, multiplestr, formatArgs(kwargs)))
  def end(self):
    self.getOut().write('</select>')

class form_option(spyceTagPlus):
  name = 'option'
  def syntax(self):
    pass
  def begin(self, text=None, value=None, selected=None, default=None, **kwargs):
    self.text = self.contextEval(text)
    value = self.contextEval(value)
    if value==None:
      valuestr = ''
    else:
      valuestr = ' value="%s"' % value
    selectTag = self.getParent('select')
    if selectTag==None:
      raise 'OPTION tag without enclosing SELECT tag'
    if value and not selected:
      selectedValues = selectTag.values
      if selectedValues==None:
        selectedValues = getFormValue(self, selectTag.varname)
      if selectedValues==None:
        selectedValues = self.contextGetModule('request').getpost(selectTag.varname)
      if selectedValues!=None and type(selectedValues) not in [type([]), type(())]:
        selectedValues = [selectedValues,]
      if selectedValues!=None:
        selected = value in selectedValues
      else:
        if default:
          selected = 1
        else:
          defaultValues = selectTag.default
          if defaultValues==None:
            defaultValues = getFormDefault(self, selectTag.varname)
          if defaultValues!=None and type(defaultValues) not in [type([]), type(())]:
            defaultValues = [defaultValues,]
          if defaultValues!=None:
            selected = value in defaultValues
    if selected:
      selectedstr = ' SELECTED'
    else:
      selectedstr = ''
    self.getOut().write('<option value="%s"%s%s>' % (
      value, selectedstr, formatArgs(kwargs)) )
    if self.text: self.buffer=1
  def body(self, contents):
    if self.text:
      self.getOut().write(self.text)
  def end(self):
    self.getOut().write('</option>')

class form(spyceTagLibrary):
  tags = [
    form_form,
    form_submit,
    form_hidden,
    form_text,
    form_password,
    form_textarea,
    form_radio,
    form_checkbox,
    form_select,
    form_option,
  ] 

